@@ -36,6 +36,20 @@ static guint agent_watch = 0;
 static gchar *agent_path = NULL;
 static gchar *agent_sender = NULL;
 
+typedef void (*agent_queue_cb)(DBusMessage *reply, void *user_data);
+
+struct agent_data {
+	struct connman_service *service;
+	DBusMessage *msg;
+	DBusPendingCall *call;
+	int timeout;
+	agent_queue_cb callback;
+	void *user_data;
+};
+
+static GList *agent_queue = NULL;
+static struct agent_data *agent_request = NULL;
+
 static void agent_free(void)
 {
 	agent_watch = 0;
@@ -45,9 +59,11 @@ static void agent_free(void)
 
 	g_free(agent_path);
 	agent_path = NULL;
+
+	__connman_agent_cancel(NULL);
 }
 
-static void agent_disconnect(DBusConnection *connection, void *data)
+static void agent_disconnect(DBusConnection *conn, void *data)
 {
 	DBG("data %p", data);
 
@@ -85,6 +101,167 @@ int __connman_agent_unregister(const char *sender, const char *path)
 	return 0;
 }
 
+static void agent_data_free(struct agent_data *data)
+{
+	if (data == NULL)
+		return;
+	if (data->service != NULL)
+		connman_service_unref(data->service);
+	if (data->msg != NULL)
+		dbus_message_unref(data->msg);
+	if (data->call != NULL)
+		dbus_pending_call_cancel(data->call);
+
+	g_free(data);
+}
+
+static void agent_receive_message(DBusPendingCall *call, void *user_data);
+
+static int agent_send_next_request(void)
+{
+	if (agent_request != NULL)
+		return -EBUSY;
+
+	if (agent_queue == NULL)
+		return 0;
+
+	agent_request = agent_queue->data;
+	agent_queue = g_list_remove(agent_queue, agent_request);
+
+	if (dbus_connection_send_with_reply(connection, agent_request->msg,
+					&agent_request->call,
+					agent_request->timeout)
+			== FALSE)
+		goto fail;
+
+	if (agent_request->call == NULL)
+		goto fail;
+
+	if (dbus_pending_call_set_notify(agent_request->call,
+					agent_receive_message, agent_request,
+					NULL) == FALSE)
+		goto fail;
+
+	dbus_message_unref(agent_request->msg);
+	agent_request->msg = NULL;
+	return 0;
+
+fail:
+	agent_data_free(agent_request);
+	agent_request = NULL;
+	return -ESRCH;
+}
+
+static int agent_send_cancel(void)
+{
+	DBusMessage *message;
+
+	if (agent_sender == NULL)
+		return 0;
+
+	message = dbus_message_new_method_call(agent_sender, agent_path,
+			CONNMAN_AGENT_INTERFACE, "Cancel");
+	if (message != NULL) {
+		dbus_message_set_no_reply(message, TRUE);
+		g_dbus_send_message(connection, message);
+		return 0;
+	}
+
+	connman_warn("Failed to send Cancel message to agent");
+	return -ESRCH;
+}
+
+static void agent_receive_message(DBusPendingCall *call, void *user_data)
+{
+	struct agent_data *queue_data = user_data;
+	DBusMessage *reply;
+
+	DBG("waiting for %p received %p", agent_request, queue_data);
+
+	if (agent_request != queue_data) {
+		connman_error("Agent callback expected %p got %p",
+				agent_request, queue_data);
+		return;
+	}
+
+	reply = dbus_pending_call_steal_reply(call);
+	dbus_pending_call_unref(call);
+	queue_data->call = NULL;
+
+	if (dbus_message_is_error(reply, "org.freedesktop.DBus.Error.Timeout")
+			== TRUE || dbus_message_is_error(reply,
+					"org.freedesktop.DBus.Error.TimedOut")
+			== TRUE) {
+		agent_send_cancel();
+	}
+
+	queue_data->callback(reply, queue_data->user_data);
+	dbus_message_unref(reply);
+
+	agent_data_free(queue_data);
+	agent_request = NULL;
+
+	agent_send_next_request();
+}
+
+static int agent_queue_message(struct connman_service *service,
+		DBusMessage *msg, int timeout,
+		agent_queue_cb callback, void *user_data)
+{
+	struct agent_data *queue_data;
+
+	if (service == NULL || callback == NULL)
+		return -EBADMSG;
+
+	queue_data = g_new0(struct agent_data, 1);
+	if (queue_data == NULL)
+		return -ENOMEM;
+
+	queue_data->service = connman_service_ref(service);
+	queue_data->msg = dbus_message_ref(msg);
+	queue_data->timeout = timeout;
+	queue_data->callback = callback;
+	queue_data->user_data = user_data;
+	agent_queue = g_list_append(agent_queue, queue_data);
+
+	return agent_send_next_request();
+}
+
+void __connman_agent_cancel(struct connman_service *service)
+{
+	GList *item, *next;
+	struct agent_data *queued_req;
+
+	DBG("service %p", service);
+
+	item = agent_queue;
+
+	while (item != NULL) {
+		next = g_list_next(item);
+		queued_req = item->data;
+
+		if (queued_req->service == service || service == NULL) {
+			agent_data_free(queued_req);
+			agent_queue = g_list_delete_link(agent_queue, item);
+		}
+
+		item = next;
+	}
+
+	if (agent_request == NULL)
+		return;
+
+	if (agent_request->service != service && service != NULL)
+		return;
+
+	agent_data_free(agent_request);
+	agent_request = NULL;
+
+	agent_send_cancel();
+
+	agent_send_next_request();
+}
+
 static connman_bool_t check_reply_has_dict(DBusMessage *reply)
 {
 	const char *signature = DBUS_TYPE_ARRAY_AS_STRING
@@ -111,7 +288,7 @@ struct request_input_reply {
 	void *user_data;
 };
 
-static void request_input_passphrase_reply(DBusPendingCall *call, void *user_data)
+static void request_input_passphrase_reply(DBusMessage *reply, void *user_data)
 {
 	struct request_input_reply *passphrase_reply = user_data;
 	connman_bool_t values_received = FALSE;
@@ -124,7 +301,6 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 	char *name = NULL;
 	int name_len = 0;
 	DBusMessageIter iter, dict;
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		error = dbus_message_get_error_name(reply);
@@ -204,9 +380,6 @@ done:
 				identity, passphrase,
 				wps, wpspin, error,
 				passphrase_reply->user_data);
-	connman_service_unref(passphrase_reply->service);
-	dbus_message_unref(reply);
-	dbus_pending_call_unref(call);
 	g_free(passphrase_reply);
 }
 
@@ -334,42 +507,66 @@ static void request_input_append_password(DBusMessageIter *iter,
 				DBUS_TYPE_STRING, &str);
 }
 
+struct previous_passphrase_data {
+	const char *passphrase;
+	const char *type;
+};
+
 static void request_input_append_previouspassphrase(DBusMessageIter *iter,
 							void *user_data)
 {
-	struct connman_service *service = user_data;
-	enum connman_service_security security;
-	const char *passphrase, *str = NULL;
-
-	passphrase = __connman_service_get_passphrase(service);
-
-	security = __connman_service_get_security(service);
-	switch (security) {
-	case CONNMAN_SERVICE_SECURITY_WEP:
-		str = "wep";
-		break;
-	case CONNMAN_SERVICE_SECURITY_PSK:
-		str  = "psk";
-		break;
-	/*
-	 * This should never happen: no passphrase is set if security is not
-	 * one of the above.*/
-	default:
-		break;
-	}
+	struct previous_passphrase_data *data = user_data;
+	const char *requirement = "informational";
 
 	connman_dbus_dict_append_basic(iter, "Type",
-				DBUS_TYPE_STRING, &str);
+				DBUS_TYPE_STRING, &data->type);
 
-	str = "informational";
 	connman_dbus_dict_append_basic(iter, "Requirement",
-				DBUS_TYPE_STRING, &str);
+				DBUS_TYPE_STRING, &requirement);
 
 	connman_dbus_dict_append_basic(iter, "Value",
-				DBUS_TYPE_STRING, &passphrase);
+				DBUS_TYPE_STRING, &data->passphrase);
 }
 
-static void request_input_login_reply(DBusPendingCall *call, void *user_data)
+static void previous_passphrase_handler(DBusMessageIter *iter,
+					struct connman_service *service)
+{
+	enum connman_service_security security;
+	struct previous_passphrase_data data;
+	struct connman_network *network;
+
+	network = __connman_service_get_network(service);
+	data.passphrase = connman_network_get_string(network, "WiFi.PinWPS");
+
+	if (connman_network_get_bool(network, "WiFi.UseWPS") == TRUE &&
+						data.passphrase != NULL) {
+		data.type = "wpspin";
+	} else {
+		data.passphrase = __connman_service_get_passphrase(service);
+		if (data.passphrase == NULL)
+			return;
+
+		security = __connman_service_get_security(service);
+		switch (security) {
+		case CONNMAN_SERVICE_SECURITY_WEP:
+			data.type = "wep";
+			break;
+		case CONNMAN_SERVICE_SECURITY_PSK:
+			data.type  = "psk";
+			break;
+		/*
+		 * This should never happen: no passphrase is set if security
+		 * is not one of the above. */
+		default:
+			break;
+		}
+	}
+
+	connman_dbus_dict_append_dict(iter, "PreviousPassphrase",
+			request_input_append_previouspassphrase, &data);
+}
+
+static void request_input_login_reply(DBusMessage *reply, void *user_data)
 {
 	struct request_input_reply *username_password_reply = user_data;
 	const char *error = NULL;
@@ -378,7 +575,6 @@ static void request_input_login_reply(DBusPendingCall *call, void *user_data)
 	char *password = NULL;
 	char *key;
 	DBusMessageIter iter, dict;
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		error = dbus_message_get_error_name(reply);
@@ -427,8 +623,6 @@ done:
 					username, password,
 					FALSE, NULL, error,
 					username_password_reply->user_data);
-	connman_service_unref(username_password_reply->service);
-	dbus_message_unref(reply);
 	g_free(username_password_reply);
 }
 
@@ -439,8 +633,8 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 	const char *path;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	DBusPendingCall *call;
 	struct request_input_reply *passphrase_reply;
+	int err;
 
 	if (service == NULL || agent_path == NULL || callback == NULL)
 		return -ESRCH;
@@ -477,10 +671,7 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 		connman_dbus_dict_append_dict(&dict, "Passphrase",
 					request_input_append_passphrase, service);
 
-		if (__connman_service_get_passphrase(service) != NULL)
-			connman_dbus_dict_append_dict(&dict, "PreviousPassphrase",
-					request_input_append_previouspassphrase,
-					service);
+		previous_passphrase_handler(&dict, service);
 	}
 
 	if (__connman_service_wps_enabled(service) == TRUE) {
@@ -496,26 +687,21 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message, &call,
-					connman_timeout_input_request())
-			== FALSE) {
-		dbus_message_unref(message);
-		g_free(passphrase_reply);
-		return -ESRCH;
-	}
-
-	if (call == NULL) {
-		dbus_message_unref(message);
-		g_free(passphrase_reply);
-		return -ESRCH;
-	}
-
-	passphrase_reply->service = connman_service_ref(service);
+	passphrase_reply->service = service;
 	passphrase_reply->callback = callback;
 	passphrase_reply->user_data = user_data;
 
-	dbus_pending_call_set_notify(call, request_input_passphrase_reply,
-				passphrase_reply, NULL);
+	err = agent_queue_message(service, message,
+			connman_timeout_input_request(),
+			request_input_passphrase_reply,
+			passphrase_reply);
+
+	if (err < 0 && err != -EBUSY) {
+		DBG("error %d sending agent message", err);
+		dbus_message_unref(message);
+		g_free(passphrase_reply);
+		return err;
+	}
 
 	dbus_message_unref(message);
 
@@ -529,8 +715,8 @@ int __connman_agent_request_login_input(struct connman_service *service,
 	const char *path;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	DBusPendingCall *call;
 	struct request_input_reply *username_password_reply;
+	int err;
 
 	if (service == NULL || agent_path == NULL || callback == NULL)
 		return -ESRCH;
@@ -563,26 +749,19 @@ int __connman_agent_request_login_input(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message, &call,
-					connman_timeout_input_request())
-			== FALSE) {
-		dbus_message_unref(message);
-		g_free(username_password_reply);
-		return -ESRCH;
-	}
-
-	if (call == NULL) {
-		dbus_message_unref(message);
-		g_free(username_password_reply);
-		return -ESRCH;
-	}
-
-	username_password_reply->service = connman_service_ref(service);
+	username_password_reply->service = service;
 	username_password_reply->callback = callback;
 	username_password_reply->user_data = user_data;
 
-	dbus_pending_call_set_notify(call, request_input_login_reply,
-						username_password_reply, NULL);
+	err = agent_queue_message(service, message,
+			connman_timeout_input_request(),
+			request_input_login_reply, username_password_reply);
+	if (err < 0 && err != -EBUSY) {
+		DBG("error %d sending agent request", err);
+		dbus_message_unref(message);
+		g_free(username_password_reply);
+		return err;
+	}
 
 	dbus_message_unref(message);
 
@@ -595,10 +774,9 @@ struct request_browser_reply_data {
 	void *user_data;
 };
 
-static void request_browser_reply(DBusPendingCall *call, void *user_data)
+static void request_browser_reply(DBusMessage *reply, void *user_data)
 {
 	struct request_browser_reply_data *browser_reply_data = user_data;
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	connman_bool_t result = FALSE;
 	const char *error = NULL;
 
@@ -612,8 +790,6 @@ static void request_browser_reply(DBusPendingCall *call, void *user_data)
 done:
 	browser_reply_data->callback(browser_reply_data->service, result,
 					error, browser_reply_data->user_data);
-	connman_service_unref(browser_reply_data->service);
-	dbus_message_unref(reply);
 	g_free(browser_reply_data);
 }
 
@@ -622,10 +798,10 @@ int __connman_agent_request_browser(struct connman_service *service,
 				const char *url, void *user_data)
 {
 	struct request_browser_reply_data *browser_reply_data;
-	DBusPendingCall *call;
 	DBusMessage *message;
 	DBusMessageIter iter;
 	const char *path;
+	int err;
 
 	if (service == NULL || agent_path == NULL || callback == NULL)
 		return -ESRCH;
@@ -652,26 +828,20 @@ int __connman_agent_request_browser(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message, &call,
-					connman_timeout_browser_launch())
-			== FALSE) {
-		dbus_message_unref(message);
-		g_free(browser_reply_data);
-		return -ESRCH;
-	}
-
-	if (call == NULL) {
-		dbus_message_unref(message);
-		g_free(browser_reply_data);
-		return -ESRCH;
-	}
-
-	browser_reply_data->service = connman_service_ref(service);
+	browser_reply_data->service = service;
 	browser_reply_data->callback = callback;
 	browser_reply_data->user_data = user_data;
 
-	dbus_pending_call_set_notify(call, request_browser_reply,
-						browser_reply_data, NULL);
+	err = agent_queue_message(service, message,
+			connman_timeout_browser_launch(),
+			request_browser_reply, browser_reply_data);
+
+	if (err < 0 && err != -EBUSY) {
+		DBG("error %d sending browser request", err);
+		dbus_message_unref(message);
+		g_free(browser_reply_data);
+		return err;
+	}
 
 	dbus_message_unref(message);
 
@@ -684,10 +854,9 @@ struct report_error_data {
 	void *user_data;
 };
 
-static void report_error_reply(DBusPendingCall *call, void *user_data)
+static void report_error_reply(DBusMessage *reply, void *user_data)
 {
 	struct report_error_data *report_error = user_data;
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	gboolean retry = FALSE;
 	const char *dbus_err;
 
@@ -701,10 +870,7 @@ static void report_error_reply(DBusPendingCall *call, void *user_data)
 
 	report_error->callback(report_error->service, retry,
 			report_error->user_data);
-	connman_service_unref(report_error->service);
 	g_free(report_error);
-	dbus_message_unref(reply);
-	dbus_pending_call_unref(call);
 }
 
 int __connman_agent_report_error(struct connman_service *service,
@@ -715,7 +881,7 @@ int __connman_agent_report_error(struct connman_service *service,
 	DBusMessageIter iter;
 	const char *path;
 	struct report_error_data *report_error;
-	DBusPendingCall *call;
+	int err;
 
 	if (service == NULL || agent_path == NULL || error == NULL ||
 		callback == NULL)
@@ -741,25 +907,20 @@ int __connman_agent_report_error(struct connman_service *service,
 		return -ENOMEM;
 	}
 
-	if (dbus_connection_send_with_reply(connection, message, &call,
-					connman_timeout_input_request())
-			== FALSE) {
-		dbus_message_unref(message);
-		g_free(report_error);
-		return -ESRCH;
-	}
-
-	if (call == NULL) {
-		dbus_message_unref(message);
-		g_free(report_error);
-		return -ESRCH;
-	}
-
-	report_error->service = connman_service_ref(service);
+	report_error->service = service;
 	report_error->callback = callback;
 	report_error->user_data = user_data;
-	dbus_pending_call_set_notify(call, report_error_reply,
-				report_error, NULL);
+
+	err = agent_queue_message(service, message,
+			connman_timeout_input_request(),
+			report_error_reply, report_error);
+	if (err < 0 && err != -EBUSY) {
+		DBG("error %d sending error request", err);
+		g_free(report_error);
+		dbus_message_unref(message);
+		return -ESRCH;
+	}
+
 	dbus_message_unref(message);
 
 	return -EINPROGRESS;
@@ -788,19 +949,16 @@ void __connman_agent_cleanup(void)
 	if (agent_watch > 0)
 		g_dbus_remove_watch(connection, agent_watch);
 
-	if (agent_path == NULL)
-		return;
+	if (agent_path != NULL) {
+		message = dbus_message_new_method_call(agent_sender, agent_path,
+				CONNMAN_AGENT_INTERFACE, "Release");
+		if (message != NULL) {
+			dbus_message_set_no_reply(message, TRUE);
+			g_dbus_send_message(connection, message);
+		}
 
-	message = dbus_message_new_method_call(agent_sender, agent_path,
-					CONNMAN_AGENT_INTERFACE, "Release");
-	if (message == NULL)
-		return;
-
-	dbus_message_set_no_reply(message, TRUE);
-
-	g_dbus_send_message(connection, message);
-
-	agent_free();
+		agent_free();
+	}
 
 	dbus_connection_unref(connection);
 }
