@@ -1,7 +1,7 @@
 /*
  *  Connection Manager
  *
- *  Copyright (C) 2007-2011  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -689,7 +689,10 @@ static int connman_iptables_insert_rule(struct connman_iptables *table,
 	if (new_entry == NULL)
 		return -EINVAL;
 
-	ret = connman_add_entry(table, new_entry, chain_head->next, builtin);
+	if (builtin == -1)
+		chain_head = chain_head->next;
+
+	ret = connman_add_entry(table, new_entry, chain_head, builtin);
 	if (ret < 0)
 		g_free(new_entry);
 
@@ -755,7 +758,7 @@ static gboolean is_same_match(struct xt_entry_match *xt_e_m1,
 	return TRUE;
 }
 
-static int connman_iptables_delete_rule(struct connman_iptables *table,
+static GList *find_existing_rule(struct connman_iptables *table,
 				struct ipt_ip *ip, char *chain_name,
 				char *target_name, struct xtables_target *xt_t,
 				struct xtables_match *xt_m,
@@ -766,24 +769,22 @@ static int connman_iptables_delete_rule(struct connman_iptables *table,
 	struct xt_entry_match *xt_e_m = NULL;
 	struct connman_iptables_entry *entry;
 	struct ipt_entry *entry_test;
-	int builtin, removed;
-
-	removed = 0;
+	int builtin;
 
 	chain_head = find_chain_head(table, chain_name);
 	if (chain_head == NULL)
-		return -EINVAL;
+		return NULL;
 
 	chain_tail = find_chain_tail(table, chain_name);
 	if (chain_tail == NULL)
-		return -EINVAL;
+		return NULL;
 
 	if (!xt_t && !xt_m)
-		return -EINVAL;
+		return NULL;
 
 	entry_test = new_rule(ip, target_name, xt_t, xt_rm);
 	if (entry_test == NULL)
-		return -EINVAL;
+		return NULL;
 
 	if (xt_t != NULL)
 		xt_e_t = ipt_get_target(entry_test);
@@ -798,7 +799,7 @@ static int connman_iptables_delete_rule(struct connman_iptables *table,
 	else
 		list = chain_head->next;
 
-	for (entry = NULL; list != chain_tail->prev; list = list->next) {
+	for (; list != chain_tail->prev; list = list->next) {
 		struct connman_iptables_entry *tmp;
 		struct ipt_entry *tmp_e;
 
@@ -826,14 +827,44 @@ static int connman_iptables_delete_rule(struct connman_iptables *table,
 				continue;
 		}
 
-		entry = tmp;
 		break;
 	}
 
-	if (entry == NULL) {
-		g_free(entry_test);
+	g_free(entry_test);
+
+	if (list != chain_tail->prev)
+		return list;
+
+	return NULL;
+}
+
+static int connman_iptables_delete_rule(struct connman_iptables *table,
+				struct ipt_ip *ip, char *chain_name,
+				char *target_name, struct xtables_target *xt_t,
+				struct xtables_match *xt_m,
+				struct xtables_rule_match *xt_rm)
+{
+	struct connman_iptables_entry *entry;
+	GList *chain_tail, *list;
+	int builtin, removed;
+
+	removed = 0;
+
+	chain_tail = find_chain_tail(table, chain_name);
+	if (chain_tail == NULL)
 		return -EINVAL;
-	}
+
+	list = find_existing_rule(table, ip, chain_name, target_name,
+							xt_t, xt_m, xt_rm);
+	if (list == NULL)
+		return -EINVAL;
+
+	entry = list->data;
+
+	if (entry == NULL)
+		return -EINVAL;
+
+	builtin = entry->builtin;
 
 	/* We have deleted a rule,
 	 * all references should be bumped accordingly */
@@ -868,6 +899,28 @@ static int connman_iptables_delete_rule(struct connman_iptables *table,
 	return 0;
 }
 
+static int connman_iptables_compare_rule(struct connman_iptables *table,
+				struct ipt_ip *ip, char *chain_name,
+				char *target_name, struct xtables_target *xt_t,
+				struct xtables_match *xt_m,
+				struct xtables_rule_match *xt_rm)
+{
+	struct connman_iptables_entry *entry;
+	GList *found;
+
+	found = find_existing_rule(table, ip, chain_name, target_name,
+							xt_t, xt_m, xt_rm);
+	if (found == NULL)
+		return -EINVAL;
+
+	entry = found->data;
+	if (entry == NULL)
+		return -EINVAL;
+
+	return 0;
+}
+
+
 static int connman_iptables_change_policy(struct connman_iptables *table,
 						char *chain_name, char *policy)
 {
@@ -1171,6 +1224,9 @@ static struct connman_iptables *connman_iptables_init(const char *table_name)
 	char *module = NULL;
 	socklen_t s;
 
+	if (table_name == NULL)
+		table_name = "filter";
+
 	if (xtables_insmod("ip_tables", NULL, TRUE) != 0)
 		goto err;
 
@@ -1238,6 +1294,7 @@ err:
 
 static struct option connman_iptables_opts[] = {
 	{.name = "append",        .has_arg = 1, .val = 'A'},
+	{.name = "compare",       .has_arg = 1, .val = 'C'},
 	{.name = "delete",        .has_arg = 1, .val = 'D'},
 	{.name = "flush-chain",   .has_arg = 1, .val = 'F'},
 	{.name = "insert",        .has_arg = 1, .val = 'I'},
@@ -1405,6 +1462,52 @@ done:
 	return xt_m;
 }
 
+static int parse_ip_and_mask(const char *str, struct in_addr *ip, struct in_addr *mask)
+{
+	char **tokens;
+	uint32_t prefixlength;
+	uint32_t tmp;
+	int err;
+
+	tokens = g_strsplit(str, "/", 2);
+	if (tokens == NULL)
+		return -1;
+
+	if (!inet_pton(AF_INET, tokens[0], ip)) {
+		err = -1;
+		goto out;
+	}
+
+	if (tokens[1] != NULL) {
+		prefixlength = strtol(tokens[1], NULL, 10);
+		if (prefixlength > 31) {
+			err = -1;
+			goto out;
+		}
+
+		tmp = ~(0xffffffff >> prefixlength);
+	} else {
+		tmp = 0xffffffff;
+	}
+
+	mask->s_addr = htonl(tmp);
+	ip->s_addr = ip->s_addr & mask->s_addr;
+	err = 0;
+out:
+	g_strfreev(tokens);
+
+	return err;
+}
+
+static struct connman_iptables *pre_load_table(char *table_name,
+					struct connman_iptables *table)
+{
+	if (table != NULL)
+		return table;
+
+	return connman_iptables_init(table_name);
+}
+
 int main(int argc, char *argv[])
 {
 	struct connman_iptables *table;
@@ -1415,8 +1518,7 @@ int main(int argc, char *argv[])
 	char *table_name, *chain, *new_chain, *match_name, *target_name;
 	char *delete_chain, *flush_chain, *policy;
 	int c, in_len, out_len;
-	gboolean dump, invert, delete, insert, delete_rule;
-	struct in_addr src, dst;
+	gboolean dump, invert, delete, insert, delete_rule, compare_rule;
 
 	xtables_init_all(&connman_iptables_globals, NFPROTO_IPV4);
 
@@ -1425,8 +1527,9 @@ int main(int argc, char *argv[])
 	delete = FALSE;
 	insert = FALSE;
 	delete_rule = FALSE;
-	table_name = chain = new_chain = match_name = target_name = NULL;
-	delete_chain = flush_chain = policy = NULL;
+	compare_rule = FALSE;
+	chain = new_chain = match_name = target_name = NULL;
+	delete_chain = flush_chain = policy = table_name = NULL;
 	memset(&ip, 0, sizeof(struct ipt_ip));
 	table = NULL;
 	xt_rm = NULL;
@@ -1436,19 +1539,29 @@ int main(int argc, char *argv[])
 	/* extension's options will generate false-positives errors */
 	opterr = 0;
 
-	while ((c = getopt_long(argc, argv, "-A:D:F:I:L::N:P:X:d:i:j:m:o:s:t:",
+	while ((c = getopt_long(argc, argv,
+				"-A:C:D:F:I:L::N:P:X:d:i:j:m:o:s:t:",
 				connman_iptables_globals.opts, NULL)) != -1) {
 		switch (c) {
 		case 'A':
-			/* It is either -A, -D or -I at once */
+			/* It is either -A, -C, -D or -I at once */
 			if (chain)
 				goto out;
 
 			chain = optarg;
 			break;
 
+		case 'C':
+			/* It is either -A, -C, -D or -I at once */
+			if (chain)
+				goto out;
+
+			chain = optarg;
+			compare_rule = TRUE;
+			break;
+
 		case 'D':
-			/* It is either -A, -D or -I at once */
+			/* It is either -A, -C, -D or -I at once */
 			if (chain)
 				goto out;
 
@@ -1461,7 +1574,7 @@ int main(int argc, char *argv[])
 			break;
 
 		case 'I':
-			/* It is either -A, -D or -I at once */
+			/* It is either -A, -C, -D or -I at once */
 			if (chain)
 				goto out;
 
@@ -1492,15 +1605,13 @@ int main(int argc, char *argv[])
 			break;
 
 		case 'd':
-			if (!inet_pton(AF_INET, optarg, &dst))
+			if (!parse_ip_and_mask(optarg, &ip.dst, &ip.dmsk))
 				break;
 
-			ip.dst = dst;
-			inet_pton(AF_INET, "255.255.255.255", &ip.dmsk);
-
 			if (invert)
 				ip.invflags |= IPT_INV_DSTIP;
 
+
 			break;
 
 		case 'i':
@@ -1519,6 +1630,11 @@ int main(int argc, char *argv[])
 
 		case 'j':
 			target_name = optarg;
+
+			table = pre_load_table(table_name, table);
+			if (table == NULL)
+				goto out;
+
 			xt_t = prepare_target(table, target_name);
 			if (xt_t == NULL)
 				goto out;
@@ -1527,6 +1643,11 @@ int main(int argc, char *argv[])
 
 		case 'm':
 			match_name = optarg;
+
+			table = pre_load_table(table_name, table);
+			if (table == NULL)
+				goto out;
+
 			xt_m = prepare_matches(table, &xt_rm, match_name);
 			if (xt_m == NULL)
 				goto out;
@@ -1548,12 +1669,9 @@ int main(int argc, char *argv[])
 			break;
 
 		case 's':
-			if (!inet_pton(AF_INET, optarg, &src))
+			if (!parse_ip_and_mask(optarg, &ip.src, &ip.smsk))
 				break;
 
-			ip.src = src;
-			inet_pton(AF_INET, "255.255.255.255", &ip.smsk);
-
 			if (invert)
 				ip.invflags |= IPT_INV_SRCIP;
 
@@ -1562,7 +1680,7 @@ int main(int argc, char *argv[])
 		case 't':
 			table_name = optarg;
 
-			table = connman_iptables_init(table_name);
+			table = pre_load_table(table_name, table);
 			if (table == NULL)
 				return -1;
 
@@ -1658,13 +1776,9 @@ int main(int argc, char *argv[])
 		xt_t->final_check(xt_t->tflags);
 #endif
 
-	if (table == NULL) {
-		table_name = "filter";
-
-		table = connman_iptables_init(table_name);
-		if (table == NULL)
-			return -1;
-	}
+	table = pre_load_table(table_name, table);
+	if (table == NULL)
+		return -1;
 
 	if (delete) {
 		if (delete_chain == NULL)
@@ -1711,6 +1825,20 @@ int main(int argc, char *argv[])
 			goto commit;
 		}
 
+		if (compare_rule == TRUE) {
+			int ret;
+
+			ret = connman_iptables_compare_rule(table, &ip,
+				chain, target_name, xt_t, xt_m, xt_rm);
+
+			if (ret == 0)
+				printf("Rule exists.\n");
+			else
+				printf("Rule does not exist.\n");
+
+			goto out;
+		}
+
 		if (delete_rule == TRUE) {
 			printf("Deleting %s to %s (match %s)\n", target_name,
 					chain, match_name);
